
// Copyright (c) 1999-2006 Regents of the University of California
//
// FFTW: Copyright (c) 2003,2006 Matteo Frigo
//       Copyright (c) 2003,2006 Massachusets Institute of Technology
//
// fft8g.[cpp,h]: Copyright (c) 1995-2001 Takya Ooura

// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the 
// Free Software Foundation; either version 2, or (at your option) any later
// version.

// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with this program; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

// In addition, as a special exception, the Regents of the University of
// California give permission to link the code of this program with libraries
// that provide specific optimized fast Fourier transform (FFT) functions
// as an alternative to FFTW and distribute a linked executable and 
// source code.  You must obey the GNU General Public License in all 
// respects for all of the code used other than the FFT library itself.  
// Any modification required to support these libraries must be distributed 
// under the terms of this license.  If you modify this program, you may extend 
// this exception to your version of the program, but you are not obligated to 
// do so. If you do not wish to do so, delete this exception statement from 
// your version.  Please be aware that FFTW is not covered by this exception, 
// therefore you may not use FFTW in any derivative work so modified without 
// permission of the authors of FFTW.
//
// $Id: analyzeFuncs_vector.cpp,v 1.1.2.6 2006/06/28 00:45:57 korpela Exp $

#include "config.h" 

#include <csignal>
#include <cstdlib>
#include <signal.h>
#include <setjmp.h>

#include "util.h"
#include "s_util.h"
#include "boinc_api.h"
#ifdef BOINC_APP_GRAPHICS
#include "reduce.h"
#include "graphics_data.h"
#include "graphics_api.h"
#ifdef DYNAMIC_GRAPHICS
#include "graphics_lib.h"
#endif
#endif
#include "diagnostics.h"

#include "sighandler.h"
#include "analyzeFuncs.h"
#include "analyzeFuncs_vector.h"

#include "hires_timer.h"
#include "asmlib.h"

BaseLineSmooth_func BaseLineSmoothFuncs[]={
            v_BaseLineSmooth,
#if defined(__x86_64__) || defined(_M_AMD64)
#endif
#ifdef USE_ALTIVEC
#endif
#ifdef USE_SSE3
#endif
#ifdef USE_SSE2
#endif
#ifdef USE_SSE
#endif
#ifdef USE_3DNOW
#endif
#ifdef USE_MMX
#endif
        };

const char * const BaseLineSmoothNames[]= {
            "v_BaseLineSmooth",
#if defined(__x86_64__) || defined(_M_AMD64)
#endif
#ifdef USE_ALTIVEC
#endif
#ifdef USE_SSE3
#endif
#ifdef USE_SSE2
#endif
#ifdef USE_SSE
#endif
#ifdef USE_3DNOW
#endif
#ifdef USE_MMX
#endif
        };

GetPowerSpectrum_func GetPowerSpectrumFuncs[]={
            v_GetPowerSpectrum,
#if defined(__x86_64__) || defined(_M_AMD64)
#endif
#ifdef USE_ALTIVEC
            v_vGetPowerSpectrum,
            v_vGetPowerSpectrumG4,
#endif
#ifdef USE_SSE3
#endif
#ifdef USE_SSE2
#endif
#ifdef USE_SSE
#endif
#ifdef USE_3DNOW
#endif
#ifdef USE_MMX
#endif
        };

const char * const GetPowerSpectrumNames[]= {
            "v_GetPowerSpectrum",
#if defined(__x86_64__) || defined(_M_AMD64)
#endif
#ifdef USE_ALTIVEC
            "v_vGetPowerSpectrum",
            "v_vGetPowerSpectrumG4",
#endif
#ifdef USE_SSE3
#endif
#ifdef USE_SSE2
#endif
#ifdef USE_SSE
#endif
#ifdef USE_3DNOW
#endif
#ifdef USE_MMX
#endif
        };

ChirpData_func ChirpDataFuncs[]={
                                    v_ChirpData,
#if defined(__x86_64__) || defined(_M_AMD64)
                                    v_vChirpData_x86_64,
#endif
#ifdef USE_ALTIVEC
                                    v_vChirpData,
                                    v_vChirpDataG4,
                                    v_vChirpDataG5,
#endif
#ifdef USE_SSE3
#endif
#ifdef USE_SSE2
#endif
#ifdef USE_SSE
#endif
#ifdef USE_3DNOW
#endif
#ifdef USE_MMX
#endif
                                };

const char * const  ChirpDataNames[]= {
                                          "v_ChirpData",
#if defined(__x86_64__) || defined(_M_AMD64)
                                          "v_vChirpData_x86_64",
#endif
#ifdef USE_ALTIVEC
                                          "v_vChirpData",
                                          "v_vChirpDataG4",
                                          "v_vChirpDataG5",
#endif
#ifdef USE_SSE3
#endif
#ifdef USE_SSE2
#endif
#ifdef USE_SSE
#endif
#ifdef USE_3DNOW
#endif
#ifdef USE_MMX
#endif
                                      };

Transpose_func TransposeFuncs[]={
                                    v_Transpose,
                                    v_Transpose2,
                                    v_Transpose4,
                                    v_Transpose8,
#if defined(__x86_64__) || defined(_M_AMD64)
#endif
#ifdef USE_ALTIVEC
                                    v_vTranspose,
#endif
#ifdef USE_SSE3
#endif
#ifdef USE_SSE2
#endif
#ifdef USE_SSE
#endif
#ifdef USE_3DNOW
#endif
#ifdef USE_MMX
#endif
                                };

const char * const TransposeNames[]= {
                                         "v_Transpose",
                                         "v_Transpose2",
                                         "v_Transpose4",
                                         "v_Transpose8",
#if defined(__x86_64__) || defined(_M_AMD64)
#endif
#ifdef USE_ALTIVEC
                                         "v_vTranspose",
#endif
#ifdef USE_SSE3
#endif
#ifdef USE_SSE2
#endif
#ifdef USE_SSE
#endif
#ifdef USE_3DNOW
#endif
#ifdef USE_MMX
#endif
                                     };

static bool do_print;

bool TestBoincSignalHandling() {
#ifdef USE_ASMLIB
    return true;
#endif
    install_sighandler();
    FORCE_FRAME_POINTER;
    if (setjmp(jb)) {
        uninstall_sighandler();
        return true;
    } else {
        // Try some illegal instructions to check the signal handling.
#ifdef __GNUC__
#if defined(__i386__) || defined (__x86_64__)
        __asm__ ("movl %cr4,%eax");
#elif defined(__APPLE__) 
         uninstall_sighandler();
         return true;
#elif defined(__ppc__) || defined(__sparc__)
__asm__ (".long %0": : "i" (0xfeedface) );
#endif
#elif defined(_MSC_VER) && ( defined(_M_IX86) || defined(_M_X64) )
#ifndef _NDEBUG
        uninstall_sighandler();
        return true;
#else
		__asm {
          __asm __emit 00fh
		  __asm __emit 020h
		  __asm __emit 0e0h
		}
#endif
#endif

    }
    uninstall_sighandler();
    return false;
}


BaseLineSmooth_func ChooseBaseLineSmooth() {
    hires_timer timer;
    BaseLineSmooth_func baseline_smooth;
    int i,j,rv;
    int NumDataPoints=1024*1024;
    double speed=1e+6,timing,accuracy;
    int best;
    double best_timing, best_accuracy;
    
    sah_complex *indata=(sah_complex *)malloc_a(NumDataPoints*sizeof(sah_complex),MEM_ALIGN);
    sah_complex *outdata=(sah_complex *)malloc_a(NumDataPoints*sizeof(sah_complex),MEM_ALIGN);
    sah_complex *save=(sah_complex *)malloc_a(NumDataPoints*sizeof(sah_complex),MEM_ALIGN);

    FORCE_FRAME_POINTER;
    if (!indata || !outdata || !save) {
        if (indata)
            free_a(indata);
        if (outdata)
            free_a(outdata);
        if (save)
            free_a(save);
        return v_BaseLineSmooth;
    }
    for (i=0;i<NumDataPoints;i++) {
        indata[i][0]=(rand()>(RAND_MAX/2))?-1:1;
        indata[i][1]=(rand()>(RAND_MAX/2))?-1:1;
    }
#ifndef USE_ASMLIB
    install_sighandler();
    for (i=0;(i*sizeof(BaseLineSmooth_func))<sizeof(BaseLineSmoothFuncs);i++) {
        if (!setjmp(jb)) {
#else
    for (i=0;(i*sizeof(BaseLineSmooth_func))<sizeof(BaseLineSmoothFuncs);i++) {
#endif
            j=0;
            timing=0;
            while ((j<10) || (timing<(3*timer.resolution()))) {
                memcpy(outdata,indata,NumDataPoints*sizeof(sah_complex));
                timer.start();
                rv=BaseLineSmoothFuncs[i](outdata,NumDataPoints,8192,32768);
                timing+=timer.stop();
#ifndef USE_ASMLIB
                if (rv) longjmp(jb,1);
#else
                if (rv) break;
#endif
                j++;
            }
            timing/=j;
            if (i==0) {
                accuracy=0;
                memcpy(save,outdata,NumDataPoints*sizeof(sah_complex));
            } else {
                accuracy=0;
                for (j=0;j<NumDataPoints;j++) {
                    accuracy+=pow(save[j][0]-outdata[j][0],2);
                    accuracy+=pow(save[j][1]-outdata[j][1],2);
                }
            }
            //fprintf(stderr,"%30s %7.5f %7.5f\n",BaseLineSmoothNames[i],timing,accuracy);
            if ((timing<speed) && (accuracy<1e-5)) {
                speed=timing;
                best=i;
                best_timing=timing;
                best_accuracy=accuracy;
                baseline_smooth=BaseLineSmoothFuncs[i];
            }
#ifndef USE_ASMLIB
        } else {
            reinstall_sighandler();
        }
    }
    uninstall_sighandler();
#else
    }
#endif
    free_a(indata);
    free_a(outdata);
    free_a(save);
    if (do_print)
      fprintf(stderr,"%30s %7.5f %7.5f\n",BaseLineSmoothNames[best],best_timing,best_accuracy);
    return baseline_smooth;
}

GetPowerSpectrum_func ChooseGetPowerSpectrum() {
    hires_timer timer;
    GetPowerSpectrum_func get_power_spectrum;
    int i,j,rv;
    double speed=1e+6,timing,accuracy;
    int NumDataPoints=128*1024;
    int best;
    double best_timing, best_accuracy;
    sah_complex *indata=(sah_complex *)malloc_a(NumDataPoints*sizeof(sah_complex),MEM_ALIGN);
    float *outdata=(float *)malloc_a(NumDataPoints*sizeof(float),MEM_ALIGN);
    float *save=(float *)malloc_a(NumDataPoints*sizeof(float),MEM_ALIGN);

    FORCE_FRAME_POINTER;
    if (!indata || !outdata || !save) {
        if (indata)
            free(indata);
        if (outdata)
            free(outdata);
        if (save)
            free(save);
        return v_GetPowerSpectrum;
    }
    for (i=0;i<NumDataPoints;i++) {
        indata[i][0]=rand()-RAND_MAX/2;
        indata[i][1]=rand()-RAND_MAX/2;
    }
#ifndef USE_ASMLIB
    install_sighandler();
    for (i=0;(i*sizeof(GetPowerSpectrum_func))<sizeof(GetPowerSpectrumFuncs);i++) {
        if (!setjmp(jb)) {
#else
    for (i=0;(i*sizeof(GetPowerSpectrum_func))<sizeof(GetPowerSpectrumFuncs);i++) {
#endif
            j=0;
            timing=0;
            while ((j<10) || (timing<(3*timer.resolution()))) {
                memset(outdata,0,NumDataPoints*sizeof(float));
                timer.start();
                rv=GetPowerSpectrumFuncs[i](indata,outdata,NumDataPoints);
                timing+=timer.stop();
#ifndef USE_ASMLIB
                if (rv) longjmp(jb,1);
#else
                if (rv) break;
#endif
                j++;
            }
            timing/=j;
            if (i==0) {
                accuracy=0;
                memcpy(save,outdata,NumDataPoints*sizeof(float));
            } else {
                accuracy=0;
                for (j=0;j<NumDataPoints;j++) {
                    accuracy+=pow(save[j]-outdata[j],2);
                }
            }
            //fprintf(stderr,"%30s %7.5f %7.5f\n",GetPowerSpectrumNames[i],timing,accuracy);
            if ((timing<speed) && (accuracy<1e-5)) {
                speed=timing;
                best=i;
                best_timing=timing;
                best_accuracy=accuracy;
                get_power_spectrum=GetPowerSpectrumFuncs[i];
            }
#ifndef USE_ASMLIB
        } else {
            reinstall_sighandler();
        }
    }
    uninstall_sighandler();
#else
    }
#endif
    free_a(indata);
    free_a(outdata);
    free_a(save);
    if (do_print)
      fprintf(stderr,"%30s %7.5f %7.5f\n",GetPowerSpectrumNames[best],best_timing,best_accuracy);
    return get_power_spectrum;
}

ChirpData_func ChooseChirpData() {
    hires_timer timer;
    ChirpData_func chirp_data;
    int i,j,rv;
    double speed=1e+6,timing,accuracy;
    int NumDataPoints=1024*1024;
    int best;
    int jj=0;
    double best_timing, best_accuracy;
    sah_complex *indata=(sah_complex *)malloc_a(NumDataPoints*sizeof(sah_complex),MEM_ALIGN);
    sah_complex *outdata=(sah_complex *)malloc_a(NumDataPoints*sizeof(sah_complex),MEM_ALIGN);
    sah_complex *save=(sah_complex *)malloc_a(NumDataPoints*sizeof(sah_complex),MEM_ALIGN);
    sah_complex *test=(sah_complex *)malloc_a(NumDataPoints*sizeof(sah_complex),MEM_ALIGN);

    FORCE_FRAME_POINTER;
    if (!indata || !outdata || !save) {
        if (indata)
            free_a(indata);
        if (outdata)
            free_a(outdata);
        if (save)
            free_a(save);
        if (test)
            free_a(test);
        fprintf(stderr,"Memory allocation failed in ChooseChirp\n");
        return v_ChirpData;
    }
    for (i=0;i<NumDataPoints;i++) {
        indata[i][0]=(rand()>(RAND_MAX/2))?-1:1;
        indata[i][1]=(rand()>(RAND_MAX/2))?-1:1;
    }
#ifndef USE_ASMLIB
    install_sighandler();
    for (i=0;(i*sizeof(ChirpData_func))<sizeof(ChirpDataFuncs);i++) {
        if (!setjmp(jb)) {
#else
    for (i=0;(i*sizeof(ChirpData_func))<sizeof(ChirpDataFuncs);i++) {
#endif
            j=0;
            timing=0;
            int ind=0;
            while ((j<10) || (timing<(3*timer.resolution()))) {
                memset(outdata,0,NumDataPoints*sizeof(sah_complex));
                timer.start();
                rv=ChirpDataFuncs[i](indata,outdata,ind,MinChirpStep*ind,NumDataPoints,2.5e+6/256.0);
                timing+=timer.stop();
#ifndef USE_ASMLIB
                if (rv) longjmp(jb,1);
#else
                if (rv) break;
#endif
                if (ind==-1) memcpy(test,outdata,NumDataPoints*sizeof(sah_complex));
                if (ind>0) {
                    ind*=-1;
                } else {
                    ind=-1*(ind-1);
                }
                j++;
            }
	    jj=j;
            timing/=j;
            if (i==0) {
                accuracy=0;
                memcpy(save,test,NumDataPoints*sizeof(sah_complex));
            } else {
                accuracy=0;
                for (j=0;j<NumDataPoints;j++) {
                    accuracy+=pow(save[j][0]-test[j][0],2);
                    accuracy+=pow(save[j][1]-test[j][1],2);
                }
            }
            //fprintf(stderr,"%30s %7.5f %7.5f\n",ChirpDataNames[i],timing,accuracy);
            if ((timing<speed) && (accuracy<5e-3)) {
                speed=timing;
                best=i;
                best_timing=timing;
                best_accuracy=accuracy;
                chirp_data=ChirpDataFuncs[i];
            }
#ifndef USE_ASMLIB
        } else {
            reinstall_sighandler();
        }
    }
    uninstall_sighandler();
#else
    }
#endif
    free_a(indata);
    free_a(outdata);
    free_a(save);
    free_a(test);
    if (do_print)
        fprintf(stderr,"%30s %7.5f %7.5f\n",ChirpDataNames[best],best_timing,best_accuracy);    
    return chirp_data;
}

Transpose_func ChooseTranspose() {
    hires_timer timer;
    Transpose_func transpose;
    int i,j,rv;
    double speed=1e+6,timing,accuracy;
    int NumDataPoints=1024*1024;
    int best;
    double best_timing, best_accuracy;    
    float *indata=(float *)malloc_a(NumDataPoints*sizeof(float),MEM_ALIGN);
    float *outdata=(float *)malloc_a(NumDataPoints*sizeof(float),MEM_ALIGN);
    float *save=(float *)malloc_a(NumDataPoints*sizeof(float),MEM_ALIGN);

    FORCE_FRAME_POINTER;
    if (!indata || !outdata || !save) {
        if (indata)
            free(indata);
        if (outdata)
            free(outdata);
        if (save)
            free(save);
        return v_Transpose;
    }
    for (i=0;i<NumDataPoints;i++) {
        indata[i]=rand();
    }
#ifndef USE_ASMLIB
    install_sighandler();
    for (i=0;(i*sizeof(Transpose_func))<sizeof(TransposeFuncs);i++) {
        if (!setjmp(jb)) {
#else
    for (i=0;(i*sizeof(Transpose_func))<sizeof(TransposeFuncs);i++) {
#endif
            j=0;
            timing=0;
            int ind=0;
            while ((j<10) || (timing<(3*timer.resolution()))) {
                memset(outdata,0,NumDataPoints*sizeof(float));
                timer.start();
                rv=TransposeFuncs[i](512,2048,indata,outdata);
                timing+=timer.stop();
#ifndef USE_ASMLIB
                if (rv) longjmp(jb,1);
#else
                if (rv) break;
#endif
                j++;
            }
            timing/=j;
            if (i==0) {
                accuracy=0;
                memcpy(save,outdata,NumDataPoints*sizeof(float));
            } else {
                accuracy=0;
                for (j=0;j<NumDataPoints;j++) {
                    accuracy+=pow(save[j]-outdata[j],2);
                }
            }
            //fprintf(stderr,"%30s %7.5f %7.5f\n",TransposeNames[i],timing,accuracy);
            if ((timing<speed) && (accuracy<1e-6)) {
                speed=timing;
                best=i;
                best_timing=timing;
                best_accuracy=accuracy;
                transpose=TransposeFuncs[i];
            }
#ifndef USE_ASMLIB
        } else {
            reinstall_sighandler();
        }
    }
    uninstall_sighandler();
#else
    }
#endif
    free_a(indata);
    free_a(outdata);
    free_a(save);
    if (do_print)
        fprintf(stderr,"%30s %7.5f %7.5f\n",TransposeNames[best],best_timing,best_accuracy);    
    return transpose;
}

void ChooseFunctions(BaseLineSmooth_func *baseline_smooth,
                     GetPowerSpectrum_func *get_power_spectrum,
                     ChirpData_func *chirp_data,
                     Transpose_func *transpose,
                     bool print_choices) {
    do_print=print_choices;
    if (do_print) {
      fprintf(stderr,"Optimal function choices:\n");
      fprintf(stderr,"-----------------------------------------------------\n");
      fprintf(stderr,"%30s %7s %8s\n","name","timing","error");
      fprintf(stderr,"-----------------------------------------------------\n");
    }
    if (TestBoincSignalHandling()) {
        *baseline_smooth=ChooseBaseLineSmooth();
        *get_power_spectrum=ChooseGetPowerSpectrum();
        *chirp_data=ChooseChirpData();
        *transpose=ChooseTranspose();
    }
    if (do_print) {
      fflush(stderr);
    }
}


